import should from 'should/as-function';
import _ from 'lodash';
import TemplateConditionModel from '../models/condition_template';
import SampleModel from '../models/sample';
import MeasurementModel from '../models/measurement';
import MaterialModel from '../models/material';
import TestHelper from "../test/helper";
import mongoose from 'mongoose';


describe('/template', () => {
  let server;
  before(done => TestHelper.before(done));
  beforeEach(done => server = TestHelper.beforeEach(server, done));
  afterEach(done => TestHelper.afterEach(server, done));
  after(done => TestHelper.after(done));

  describe('/template/condition', () => {
    describe('GET /template/conditions', () => {
      it('returns all condition templates', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/conditions',
          auth: {basic: 'janedoe'},
          httpStatus: 200
        }).end((err, res) => {
          if (err) return done(err);
          const json = require('../test/db.json');
          should(res.body).have.lengthOf(json.collections.condition_templates.length);
          should(res.body).matchEach(condition => {
            should(condition).have.only.keys('_id', 'name', 'version', 'first_id', 'parameters');
            should(condition).have.property('_id').be.type('string');
            should(condition).have.property('name').be.type('string');
            should(condition).have.property('version').be.type('number');
            should(condition).have.property('first_id').be.type('string');
            should(condition.parameters).matchEach(number => {
              should(number).have.only.keys('name', 'range');
              should(number).have.property('name').be.type('string');
              should(number).have.property('range').be.type('object');
            });
          });
          done();
        });
      });
      it('rejects an API key', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/conditions',
          auth: {key: 'janedoe'},
          httpStatus: 401
        });
      });
      it('rejects unauthorized requests', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/conditions',
          httpStatus: 401
        });
      });
    });

    describe('GET /template/condition/{id}', () => {
      it('returns the right condition template', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'janedoe'},
          httpStatus: 200,
          res: {_id: '200000000000000000000001', name: 'heat treatment', version: 1, first_id: '200000000000000000000001', parameters: [{name: 'material', range: {values: ['copper', 'hot air']}}, {name: 'weeks', range: {min: 1, max: 10, required: true}}]}
        });
      });
      it('rejects an API key', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/condition/200000000000000000000001',
          auth: {key: 'janedoe'},
          httpStatus: 401
        });
      });
      it('rejects an unknown id', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/condition/000000000000000000000001',
          auth: {basic: 'janedoe'},
          httpStatus: 404
        });
      });
      it('rejects unauthorized requests', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/condition/200000000000000000000001',
          httpStatus: 401
        });
      });
    });

    describe('PUT /template/condition/{id}', () => {
      it('returns the right condition template', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {},
          res: {_id: '200000000000000000000001', name: 'heat treatment', version: 1, first_id: '200000000000000000000001', parameters: [{name: 'material', range: {values: ['copper', 'hot air']}}, {name: 'weeks', range: {min: 1, max: 10, required: true}}]}
        });
      });
      it('keeps unchanged properties', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'heat treatment', parameters: [{name: 'material', range: {values: ['copper', 'hot air']}}, {name: 'weeks', range: {min: 1, max: 10, required: true}}]},
          res: {_id: '200000000000000000000001', name: 'heat treatment', version: 1, first_id: '200000000000000000000001', parameters: [{name: 'material', range: {values: ['copper', 'hot air']}}, {name: 'weeks', range: {min: 1, max: 10, required: true}}]}
        });
      });
      it('keeps only one unchanged property', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'heat treatment'},
          res: {_id: '200000000000000000000001', name: 'heat treatment', version: 1, first_id: '200000000000000000000001', parameters: [{name: 'material', range: {values: ['copper', 'hot air']}}, {name: 'weeks', range: {min: 1, max: 10, required: true}}]}
        });
      });
      it('changes the given properties', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {min: 1}}]}
        }).end((err, res) => {
            if (err) return done(err);
            TemplateConditionModel.findById(res.body._id).lean().exec((err, data:any) => {
              if (err) return done(err);
              should(data).have.only.keys('_id', 'first_id', 'name', 'version', 'parameters', '__v');
              should(data.first_id.toString()).be.eql('200000000000000000000001');
              should(data).have.property('name', 'heat aging');
              should(data).have.property('version', 2);
              should(data.first_id.toString()).be.eql('200000000000000000000001');
              should(data).have.property('parameters').have.lengthOf(1);
              should(data.parameters[0]).have.property('name', 'time');
              should(data.parameters[0]).have.property('range');
              should(data.parameters[0].range).have.property('min', 1);
              done();
            });
          });
      });
      it('renames all occurrences instead of creating a new version when only the parameter name is changed', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200, req: {name: 'heat treatment', parameters: [{name: 'treatmentMaterial', range: {values: ['copper', 'hot air']}}, {name: 'weeks', range: {min: 1, max: 10, required: true}}]}
        }).end((err, res) => {
          if (err) return done(err);
          should(res.body).be.eql({_id: '200000000000000000000001', name: 'heat treatment', version: 1, first_id: '200000000000000000000001', parameters: [{name: 'treatmentMaterial', range: {values: ['copper', 'hot air']}}, {name: 'weeks', range: {min: 1, max: 10, required: true}}]});
          SampleModel.find({'condition.condition_template': mongoose.Types.ObjectId('200000000000000000000001')}).lean().exec((err, data:any) => {
            if (err) return done(err);
            should(data).matchEach(sample => {
              should(sample.condition).have.only.keys('treatmentMaterial', 'weeks', 'condition_template');
            });
            done();
          });
        });
      });
      it('creates a changelog', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {min: 1}}]},
          log: {
            collection: 'condition_templates',
            dataAdd: {
              first_id: '200000000000000000000001',
              version: 2
            }
          }
        });
      });
      it('does not increase the version on name change', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'heat aging'}
        }).end((err, res) => {
          if (err) return done(err);
          TemplateConditionModel.findById(res.body._id).lean().exec((err, data:any) => {
            if (err) return done(err);
            should(data).have.only.keys('_id', 'first_id', 'name', 'version', 'parameters', '__v');
            should(data.first_id.toString()).be.eql('200000000000000000000001');
            should(data).have.property('name', 'heat aging');
            should(data).have.property('version', 1);
            should(data).have.property('parameters').have.lengthOf(2);
            should(data.parameters[0]).have.property('name', 'material');
            should(data.parameters[1]).have.property('name', 'weeks');
            done();
          });
        });
      });
      it('does not increase the version on name change when property ranges stayed the same', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'heat aging', parameters: [{name: 'material', range: {values: ['copper', 'hot air']}}, {name: 'duration', range: {min: 1, max: 10, required: true}}]}
        }).end((err, res) => {
          if (err) return done(err);
          TemplateConditionModel.findById(res.body._id).lean().exec((err, data:any) => {
            if (err) return done(err);
            should(data).have.only.keys('_id', 'first_id', 'name', 'version', 'parameters', '__v');
            should(data.first_id.toString()).be.eql('200000000000000000000001');
            should(data).have.property('name', 'heat aging');
            should(data).have.property('version', 1);
            should(data).have.property('parameters').have.lengthOf(2);
            should(data.parameters[0]).have.property('name', 'material');
            should(data.parameters[1]).have.property('name', 'duration');
            done();
          });
        });
      });
      it('supports values ranges', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {parameters: [{name: 'time', range: {values: [1, 2, 5]}}]}
        }).end((err, res) => {
          if (err) return done(err);
          should(_.omit(res.body, '_id')).be.eql({name: 'heat treatment', version: 2, first_id: '200000000000000000000001', parameters: [{name: 'time', range: {values: [1, 2, 5]}}]});
          done();
        });
      });
      it('supports min max ranges', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {parameters: [{name: 'time', range: {min: 1, max: 11}}]}
        }).end((err, res) => {
          if (err) return done(err);
          should(_.omit(res.body, '_id')).be.eql({name: 'heat treatment', version: 2, first_id: '200000000000000000000001', parameters: [{name: 'time', range: {min: 1, max: 11}}]});
          done();
        });
      });
      it('supports array type ranges', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {parameters: [{name: 'time', range: {type: 'array'}}]}
        }).end((err, res) => {
          if (err) return done(err);
          should(_.omit(res.body, '_id')).be.eql({name: 'heat treatment', version: 2, first_id: '200000000000000000000001', parameters: [{name: 'time', range: {type: 'array'}}]});
          done();
        });
      });
      it('supports required ranges', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {parameters: [{name: 'time', range: {required: true}}]}
        }).end((err, res) => {
          if (err) return done(err);
          should(_.omit(res.body, '_id')).be.eql({name: 'heat treatment', version: 2, first_id: '200000000000000000000001', parameters: [{name: 'time', range: {required: true}}]});
          done();
        });
      });
      it('supports empty ranges', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {parameters: [{name: 'time', range: {}}]}
        }).end((err, res) => {
          if (err) return done(err);
          should(_.omit(res.body, '_id')).be.eql({name: 'heat treatment', version: 2, first_id: '200000000000000000000001', parameters: [{name: 'time', range: {}}]});
          done();
        });
      });
      it('rejects `condition_template` as parameter name', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {parameters: [{name: 'condition_template', range: {}}]},
          res: {status: 'Invalid body format', details: '"parameters[0].name" contains an invalid value'}
        });
      });
      it('rejects not specified parameters', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {name: 'heat treatment', parameters: [{name: 'material', range: {xx: 5}}]},
          res: {status: 'Invalid body format', details: '"parameters[0].range.xx" is not allowed'}
        });
      });
      it('rejects an invalid id', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/2000000000h0000000000001',
          auth: {basic: 'admin'},
          httpStatus: 404,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {min: 1}}]}
        });
      });
      it('rejects an unknown id', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/000000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 404,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {min: 1}}]}
        });
      });
      it('rejects an API key', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {key: 'admin'},
          httpStatus: 401,
          req: {}
        });
      });
      it('rejects requests from a write user', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          auth: {basic: 'janedoe'},
          httpStatus: 403,
          req: {}
        });
      });
      it('rejects unauthorized requests', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/condition/200000000000000000000001',
          httpStatus: 401,
          req: {}
        });
      });
    });

    describe('POST /template/condition/new', () => {
      it('returns the right condition template', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'heat treatment3', parameters: [{name: 'material', range: {values: ['copper']}}]}
        }).end((err, res) => {
          if (err) return done(err);
          should(res.body).have.only.keys('_id', 'name', 'version', 'first_id', 'parameters');
          should(res.body).have.property('name', 'heat treatment3');
          should(res.body).have.property('version', 1);
          should(res.body._id).be.eql(res.body.first_id);
          should(res.body).have.property('parameters').have.lengthOf(1);
          should(res.body.parameters[0]).have.property('name', 'material');
          should(res.body.parameters[0]).have.property('range');
          should(res.body.parameters[0].range).have.property('values');
          should(res.body.parameters[0].range.values[0]).be.eql('copper');
          done();
        });
      });
      it('stores the template', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {min: 1}}]}
        }).end((err, res) => {
          if (err) return done(err);
          TemplateConditionModel.findById(res.body._id).lean().exec((err, data:any) => {
            if (err) return done(err);
            should(data).have.only.keys('_id', 'first_id', 'name', 'version', 'parameters', '__v');
            should(data.first_id.toString()).be.eql(data._id.toString());
            should(data).have.property('name', 'heat aging');
            should(data).have.property('version', 1);
            should(res.body._id).be.eql(res.body.first_id);
            should(data).have.property('parameters').have.lengthOf(1);
            should(data.parameters[0]).have.property('name', 'time');
            should(data.parameters[0]).have.property('range');
            should(data.parameters[0].range).have.property('min', 1);
            done();
          });
        });
      });
      it('creates a changelog', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {min: 1}}]},
          log: {
            collection: 'condition_templates',
            dataAdd: {version: 1},
            dataIgn: ['first_id']
          }
        });
      });
      it('rejects a missing name', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {parameters: [{name: 'time', range: {min: 1}}]},
          res: {status: 'Invalid body format', details: '"name" is required'}
        });
      });
      it('rejects `condition_template` as parameter name', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {name: 'heat aging', parameters: [{name: 'condition_template', range: {min: 1}}]},
          res: {status: 'Invalid body format', details: '"parameters[0].name" contains an invalid value'}
        });
      });
      it('rejects a number prefix', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {name: 'heat aging', number_prefix: 'C', parameters: [{name: 'time', range: {min: 1}}]},
          res: {status: 'Invalid body format', details: '"number_prefix" is not allowed'}
        });
      });
      it('rejects missing parameters', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {name: 'heat aging'},
          res: {status: 'Invalid body format', details: '"parameters" is required'}
        });
      });
      it('rejects a missing parameter name', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {name: 'heat aging', parameters: [{range: {min: 1}}]},
          res: {status: 'Invalid body format', details: '"parameters[0].name" is required'}
        });
      });
      it('rejects a missing parameter range', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {name: 'heat aging', parameters: [{name: 'time'}]},
          res: {status: 'Invalid body format', details: '"parameters[0].range" is required'}
        });
      });
      it('rejects an invalid parameter range property', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {xx: 1}}]},
          res: {status: 'Invalid body format', details: '"parameters[0].range.xx" is not allowed'}
        });
      });
      it('rejects wrong properties', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'admin'},
          httpStatus: 400,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {}}], xx: 33},
          res: {status: 'Invalid body format', details: '"xx" is not allowed'}
        });
      });
      it('rejects an API key', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {key: 'admin'},
          httpStatus: 401,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {min: 1}}]}
        });
      });
      it('rejects requests from a write user', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          auth: {basic: 'janedoe'},
          httpStatus: 403,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {min: 1}}]}
        });
      });
      it('rejects unauthorized requests', done => {
        TestHelper.request(server, done, {
          method: 'post',
          url: '/template/condition/new',
          httpStatus: 401,
          req: {name: 'heat aging', parameters: [{name: 'time', range: {min: 1}}]}
        });
      });
    });
  });

  describe('/template/measurement', () => {
    describe('GET /template/measurements', () => {
      it('returns all measurement templates', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/measurements',
          auth: {basic: 'janedoe'},
          httpStatus: 200
        }).end((err, res) => {
          if (err) return done(err);
          const json = require('../test/db.json');
          should(res.body).have.lengthOf(json.collections.measurement_templates.length);
          should(res.body).matchEach(measurement => {
            should(measurement).have.only.keys('_id', 'name', 'version', 'first_id', 'parameters');
            should(measurement).have.property('_id').be.type('string');
            should(measurement).have.property('first_id').be.type('string');
            should(measurement).have.property('name').be.type('string');
            should(measurement).have.property('version').be.type('number');
            should(measurement.parameters).matchEach(number => {
              should(number).have.only.keys('name', 'range');
              should(number).have.property('name').be.type('string');
              should(number).have.property('range').be.type('object');
            });
          });
          done();
        });
      });
      it('rejects an API key', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/measurements',
          auth: {key: 'janedoe'},
          httpStatus: 401
        });
      });
      it('rejects unauthorized requests', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/measurements',
          httpStatus: 401
        });
      });
    });
    describe('PUT /template/measurement/{id}', () => {
      it('renames all occurrences instead of creating a new version when only the parameter name is changed', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/measurement/300000000000000000000001',
          auth: {basic: 'admin'},
          httpStatus: 200, req: {name: 'spectrum', parameters: [{name: 'spectrumValues', range: {type: 'array'}}, {name: 'device', range: {}}, {name: 'filename', range: {}}]}
        }).end((err, res) => {
          if (err) return done(err);
          should(res.body).be.eql({_id: '300000000000000000000001', name: 'spectrum', version: 1, first_id: '300000000000000000000001', parameters: [{name: 'spectrumValues', range: {type: 'array'}}, {name: 'device', range: {}}, {name: 'filename', range: {}}]});
          MeasurementModel.find({'measurement_template': mongoose.Types.ObjectId('300000000000000000000001')}).lean().exec((err, data:any) => {
            if (err) return done(err);
            should(data).matchEach(measurement => {
              should(measurement.values).have.only.keys('spectrumValues', 'device', 'filename');
            });
            done();
          });
        });
      });
    });
    // other methods should be covered by condition tests
  });

  describe('/template/material', () => {
    describe('GET /template/materials', () => {
      it('returns all material templates', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/materials',
          auth: {basic: 'janedoe'},
          httpStatus: 200
        }).end((err, res) => {
          if (err) return done(err);
          const json = require('../test/db.json');
          should(res.body).have.lengthOf(json.collections.material_templates.length);
          should(res.body).matchEach(measurement => {
            should(measurement).have.only.keys('_id', 'name', 'version', 'first_id', 'parameters');
            should(measurement).have.property('_id').be.type('string');
            should(measurement).have.property('name').be.type('string');
            should(measurement).have.property('version').be.type('number');
            should(measurement.parameters).matchEach(number => {
              should(number).have.only.keys('name', 'range');
              should(number).have.property('name').be.type('string');
              should(number).have.property('range').be.type('object');
            });
          });
          done();
        });
      });
      it('rejects an API key', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/materials',
          auth: {key: 'janedoe'},
          httpStatus: 401
        });
      });
      it('rejects unauthorized requests', done => {
        TestHelper.request(server, done, {
          method: 'get',
          url: '/template/materials',
          httpStatus: 401
        });
      });
    });
    describe('PUT /template/material/{id}', () => {
      it('renames all occurrences instead of creating a new version when only the parameter name is changed', done => {
        TestHelper.request(server, done, {
          method: 'put',
          url: '/template/material/130000000000000000000003',
          auth: {basic: 'admin'},
          httpStatus: 200,
          req: {name: 'plastic', parameters: [ {name: 'glassfiber', range: {min: 0, max: 100, required: true}}, {name: 'carbonfiber', range: {min: 0, max: 100, required: true}}, {name: 'mineral', range: {min: 0, max: 100, required: true}}]}
        }).end((err, res) => {
          if (err) return done(err);
          should(res.body).be.eql({_id: '130000000000000000000003', name: 'plastic', version: 2, first_id: '130000000000000000000001', parameters: [ {name: 'glassfiber', range: {min: 0, max: 100, required: true}}, {name: 'carbonfiber', range: {min: 0, max: 100, required: true}}, {name: 'mineral', range: {min: 0, max: 100, required: true}}]});
          MaterialModel.find({'properties': mongoose.Types.ObjectId('130000000000000000000003')}).lean().exec((err, data:any) => {
            if (err) return done(err);
            should(data).matchEach(material => {
              should(material.parameters).have.only.keys('glassfiber', 'carbonfiber', 'mineral', 'material_template');
            });
            done();
          });
        });
      });
    });
    // other methods should be covered by condition tests
  });
});